/* 
 * sdbInit.c --
 *
 *  This file contains the top-level command and the glue code
 *  to the Sdb-tcl routines.  
 *
 * Copyright (c) 1995 Tom Moore.  All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice appear in all copies of this software.
 * 
 */

#include "sdbInt.h"

Tcl_HashTable           sdb_HashTable;

/* 
 * Forward declarations for procedures defined in this file:
 */

#ifndef __STDC__
static int              SdbCmd();
static int              SdbWidget();
static void             SdbDeleteWidget();
#else
static Tcl_CmdProc      SdbCmd;
static Tcl_CmdProc      SdbWidget;
static Tcl_CmdDeleteProc SdbDeleteWidget;
#endif

static int              SdbGetRecord _ANSI_ARGS_((Tcl_Interp *interp, 
			       SdbFile *sdbPtr, int rec));
static int              SdbGetCodeList _ANSI_ARGS_((Tcl_Interp *interp, 
			       SdbFile *sdbPtr,SdbVariable *varPtr));
static int              SdbGetNumberRange _ANSI_ARGS_((Tcl_Interp *interp, 
			       SdbFile *sdbPtr,SdbVariable *varPtr));
static int              SdbPutRecord _ANSI_ARGS_((Tcl_Interp *interp, 
			       SdbFile *sdbPtr, int rec, char *list));

int
Sdb_Init(interp)
     Tcl_Interp             *interp;
{

    /* Add the Sdb command definition to the interpretor */

    Tcl_CreateCommand(interp, "sdb", SdbCmd,
            (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

    Tcl_InitHashTable(&sdb_HashTable, TCL_STRING_KEYS);

    return TCL_OK;
}

static int
SdbWidget(clientData, interp, argc, argv)
     ClientData              clientData;  /* Information about database. */
     Tcl_Interp             *interp;   /* Current interpreter. */
     int                     argc;     /* Number of arguments. */
     char                  **argv;     /* Argument strings. */
{
    SdbFile                *sdbPtr = (SdbFile *) clientData;
    SdbVariable            *varPtr = sdbPtr->varPtr;
    unsigned int            length;
    int                     errcode;
    char                    c,
                            out[256];

    if (argc < 2) {
        Tcl_AppendResult(interp, "wrong # args: should be \"",
                argv[0], " option ?arg arg ...?\"", (char *) NULL);
        return TCL_ERROR;
    }
    c = argv[1][0];
    length = strlen(argv[1]);
    if ((c == 'c') && (strncmp(argv[1], "close", length) == 0)) {
        return (Tcl_DeleteCommand(interp, sdbPtr->dbname));
    } else if ((c == 'i') && (strncmp(argv[1], "info", length) == 0)
            && (length >= 3)) {

        if (argc < 3) {
            Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
                    " info option...\"", (char *) NULL);
            return TCL_ERROR;
        } else {
            int                     i,
                                    first = 1;
            unsigned int            l;
            char                    k;

            for (i = 2; i < argc; i++) {
                l = strlen(argv[i]);
                k = argv[i][0];

                if (!first)
                    Tcl_AppendResult(interp, " ", (char *) NULL);
                first = 0;

                if ((k == 'n') && (strncmp(argv[i], "num_records", l) == 0) &&
                        l > 4) {
                    sprintf(out, "%d", sdbPtr->numRec);
                    Tcl_AppendResult(interp, out, (char *) NULL);
                    return TCL_OK;
                }
                if ((k == 'n') && (strncmp(argv[i], "num_output", l) == 0) &&
                        l > 4) {
                    sprintf(out, "%d", sdbPtr->numOutput);
                    Tcl_AppendResult(interp, out, (char *) NULL);
                    return TCL_OK;
                }
                if ((k == 'n') && (strncmp(argv[i], "num_selected", l) == 0) &&
                        l > 4) {
                    sprintf(out, "%d", sdbPtr->numSelect);
                    Tcl_AppendResult(interp, out, (char *) NULL);
                    return TCL_OK;
                } else if ((k == 'n') && (strncmp(argv[i], "name", l) == 0) &&
                        l > 1) {
                    Tcl_AppendResult(interp, sdbPtr->filename, (char *) NULL);
                    return TCL_OK;
                } else if ((k == 'e') && (strncmp(argv[i], "eof", l) == 0) &&
                        l > 1) {
                    Tcl_AppendResult(interp,
                            (sdbPtr->position >= sdbPtr->numRec - 1) ?
                            "1" : "0", (char *) NULL);
                    return TCL_OK;
                } else if ((k == 'e') && 
			  (strncmp(argv[i], "errors", l) == 0) &&
			   l > 1) {
		    return(SdbErrorCheck(sdbPtr));
                } else if ((k == 'n') && (strncmp(argv[i], "num_items", l) == 0) &&
                        l > 4) {
                    int                     m = 0;

                    for (varPtr = sdbPtr->varPtr; varPtr; varPtr = varPtr->next)
                        if (varPtr->id)
                            m++;
                    sprintf(out, "%d", m);
                    Tcl_AppendResult(interp, out, (char *) NULL);
                    return TCL_OK;
                } else if ((k == 'c') && (strncmp(argv[i], "current", l)) == 0) {
                    sprintf(out, "%d", sdbPtr->position);
                    Tcl_AppendResult(interp, out, (char *) NULL);
                    return TCL_OK;
                } else if ((k == 'i') && (strncmp(argv[i], "items", l)) == 0
                        && l == 5) {
                    for (varPtr = sdbPtr->varPtr; varPtr; varPtr = varPtr->next) {
                        if (varPtr->id)
                            Tcl_AppendResult(interp, varPtr->id, " ", (char *) NULL);
                    }
                    return TCL_OK;
                } else if ((k == 'c') && (strncmp(argv[i], "codes", l)) == 0 &&
                        l == 5) {
                    if (argc == i + 1) {
                        Tcl_SetResult(interp, "\"info codes\" must specify item name",
                                TCL_STATIC);
                        return TCL_ERROR;
                    }
                    i++;
                    if ((varPtr = SdbFound(sdbPtr->varPtr, argv[i]))) {
                        if (varPtr->flag & SDB_NUMBER) {
                            Tcl_AppendResult(interp, "variable ", argv[i], " is numeric",
                                    NULL);
                            return TCL_ERROR;
                        } else if (varPtr->flag & SDB_CONSTANT) {
                            Tcl_AppendResult(interp, varPtr->u.code, NULL);
                            return TCL_OK;
                        } else {
                            return (SdbGetCodeList(interp, sdbPtr, varPtr));
                        }
                    } else {
                        Tcl_AppendResult(interp, "variable \"", argv[i],
                                "\" not found", (char *) NULL);
                        return TCL_ERROR;
                    }
                } else if (strcmp(argv[i], "range") == 0) {
                    if (argc == i + 1) {
                        Tcl_SetResult(interp, "\"info range\" must specify item name",
                                TCL_STATIC);
                        return TCL_ERROR;
                    }
                    i++;
                    if ((varPtr = SdbFound(varPtr, argv[i]))) {
                        if (varPtr->flag & SDB_CODE) {
                            Tcl_AppendResult(interp, "variable ", argv[i], " is an alpha code",
                                    NULL);
                            return TCL_ERROR;
                        } else if (varPtr->flag & SDB_CONSTANT) {
                            char                    temp[64];

                            sprintf(temp, "%-.15g", SdbGetNumber(sdbPtr, varPtr, 1));
                            Tcl_AppendResult(interp, temp, " ", temp, NULL);
                            return TCL_OK;
                        } else {
                            return (SdbGetNumberRange(interp, sdbPtr, varPtr));
                        }
                    } else {
                        Tcl_AppendResult(interp, "variable \"", argv[i],
                                "\" not found", (char *) NULL);
                        return TCL_ERROR;
                    }
                } else if ((k == 'i') && (strncmp(argv[i], "item", l)) == 0 &&
                        l == 4) {
                    if (argc == i + 1) {
                        Tcl_SetResult(interp, "\"info item\" must specify item name",
                                TCL_STATIC);
                        return TCL_ERROR;
                    }
                    i++;
                    for (varPtr = sdbPtr->varPtr; varPtr; varPtr = varPtr->next) {
                        if (varPtr->id && strcmp(varPtr->id, argv[i]) == 0) {
                            if (varPtr->flag & SDB_CONSTANT) {
                                if (varPtr->flag & SDB_NUMBER) {
                                    char                    temp[50];

                                    sprintf(temp, "%-.15g", varPtr->u.value);
                                    Tcl_AppendResult(interp, "CONSTANT NUMERIC ", temp, NULL);
                                } else {
                                    Tcl_AppendResult(interp, "CONSTANT CODE ",
                                            varPtr->u.code, NULL);
                                }
                            } else if (varPtr->flag & SDB_ALLOCATED) {
                                Tcl_AppendResult(interp, "ALLOCATED ",
                                        (varPtr->flag & SDB_NUMBER) ? "NUMERIC" : "CODE",
                                        NULL);
                            } else if (varPtr->flag & SDB_EXPRESSION) {
                                Tcl_AppendResult(interp, "EXPRESSION ",
                                        (varPtr->flag & SDB_NUMBER) ? "NUMERIC " : "CODE ",
                                        NULL);
                                Tcl_AppendResult(interp, "{", NULL);
                                SdbPrettyPrint(interp, varPtr, varPtr->u.expr[1]);
                                Tcl_AppendResult(interp, "} {", NULL);
                                SdbPrettyPrint(interp, varPtr, varPtr->u.expr[0]);
                                Tcl_AppendResult(interp, "}", NULL);
                            } else if (varPtr->flag & SDB_LOOKUP) {
                                Tcl_AppendResult(interp, "LOOKUP ",
                                        (varPtr->flag & SDB_NUMBER) ? "NUMERIC" : "CODE",
                                        NULL);
                            } else {
                                Tcl_AppendResult(interp, "Unknown?", NULL);
                            }
                            if (varPtr->flag & SDB_OUTPUT) {
                                Tcl_AppendResult(interp, " OUTPUT", NULL);
                            }
                            if (varPtr->flag & SDB_INDEXED) {
                                Tcl_AppendResult(interp, " INDEXED", NULL);
                            }
                            if (varPtr->flag & SDB_MUTABLE) {
                                Tcl_AppendResult(interp, " MUTABLE", NULL);
                            }
                            goto go_it;
                        }
                    }
                    Tcl_AppendResult(interp, "variable \"", argv[i],
                            "\" not found", (char *) NULL);
                    return TCL_ERROR;
                  go_it:;
                } else {
                    Tcl_AppendResult(interp, "bad option \"", argv[i],
                            "\": should be codes, current, eof, errors, item, ",
                            "items, name, ",
                            "num_items, num_output, num_records ",
                            "or num_selected",
                            (char *) NULL);
                    return TCL_ERROR;
                }
            }
            return TCL_OK;
        }
    } else if ((c == 'd') && (strncmp(argv[1], "define", length)) == 0) {
        if (argc != 4) {
            Tcl_SetResult(interp, "wrong number of arguments", TCL_STATIC);
            return TCL_ERROR;
        } else
            return (SdbDefine(interp, sdbPtr, argv[2], argv[3]));
    } else if ((c == 'a') && (strncmp(argv[1], "arrange", length)) == 0) {
        if (argc >= 5) {
            int                     toggle;

            if (strcmp(argv[2], "before") == 0)
                toggle = 0;
            else if (strcmp(argv[2], "after") == 0)
                toggle = 1;
            else {
                Tcl_AppendResult(interp, "invalid argument \"", argv[2],
                        "\", should be \"before\" or \"after\"", NULL);
                return TCL_ERROR;
            }
            SdbArrange(sdbPtr, argc - 3, &argv[3], toggle);
            SdbRebuildOutput(sdbPtr);
        } else {
            Tcl_SetResult(interp, "wrong number of arguments", TCL_STATIC);
            return TCL_ERROR;
        }
    } else if ((c == 'o') && (strncmp(argv[1], "output", length)) == 0) {
        int                     j;

        if (argc == 2) {
            for (varPtr = sdbPtr->varPtr; varPtr; varPtr = varPtr->next) {
                if (varPtr->flag & SDB_OUTPUT)
                    Tcl_AppendResult(interp, varPtr->id, " ", (char *) NULL);
            }
            return TCL_OK;
        } else if (argc > 3) {
            int                     toggle;

            if (strcmp(argv[2], "on") == 0)
                toggle = 1;
            else if (strcmp(argv[2], "off") == 0)
                toggle = 0;
            else {
                Tcl_AppendResult(interp, "invalid argument \"", argv[2],
                        "\", should be \"on\" or \"off\"", NULL);
                return TCL_ERROR;
            }
            for (j = 3; j < argc; j++) {
                if (!(varPtr = SdbFound(sdbPtr->varPtr, argv[j]))) {
                    Tcl_AppendResult(interp, "variable \"", argv[j],
                            "\" not found", NULL);
                    return TCL_ERROR;
                } else {
                    if (toggle)
                        varPtr->flag |= SDB_OUTPUT;
                    else
                        varPtr->flag &= ~SDB_OUTPUT;
                }
            }
            SdbRebuildOutput(sdbPtr);
        } else {
            Tcl_SetResult(interp, "wrong number of arguments", TCL_STATIC);
            return TCL_ERROR;
        }
    } else if (strcmp(argv[1], "format") == 0) {
        if (argc < 3 || argc > 4) {
            Tcl_SetResult(interp, "wrong # args: \"format item ?format?\"",
                    TCL_STATIC);
            return TCL_ERROR;
        }
        for (varPtr = sdbPtr->varPtr; varPtr; varPtr = varPtr->next) {
            if (varPtr->id && strcmp(varPtr->id, argv[2]) == 0) {
                if (argc == 3) {
                    Tcl_AppendResult(interp, varPtr->format, NULL);
                    return TCL_OK;
                } else {
                    if (varPtr->format)
                        ckfree(varPtr->format);
                    varPtr->format = strdup(argv[3]);
                    Tcl_AppendResult(interp, argv[3], NULL);
                    return TCL_OK;
                }
            }
        }
        Tcl_AppendResult(interp, "variable \"", argv[2],
                "\" not found", (char *) NULL);
        return TCL_ERROR;
    } else if ((c == 's') && (strncmp(argv[1], "select", length)) == 0) {
        if (argc < 3) {
            Tcl_SetResult(interp, "too few arguments", TCL_STATIC);
            return TCL_ERROR;
        } else
            return (SdbSelect(interp, sdbPtr, argc - 2, &argv[2]));
    } else if ((c == 'c') && (strncmp(argv[1], "calculate", length)) == 0) {
        if (argc != 3) {
            Tcl_SetResult(interp, "wrong number of arguments", TCL_STATIC);
            return TCL_ERROR;
        } else
            return (SdbCalculate(interp, sdbPtr, argv[2]));
    } else if ((c == 'd') && (strncmp(argv[1], "delete", length)) == 0) {
        if (argc != 3) {
            Tcl_SetResult(interp, "wrong number of arguments", TCL_STATIC);
            return TCL_ERROR;
        } else
            return (SdbDelete(interp, sdbPtr, argv[2], 0));
    } else if ((c == 'j') && (strncmp(argv[1], "join", length)) == 0) {
        if (argc < 2) {
            Tcl_SetResult(interp, "wrong number of arguments", TCL_STATIC);
            return TCL_ERROR;
        } else
            return (SdbJoinCmd(interp, sdbPtr, argc - 2, argv + 2));
    } else if ((c == 'r') && (strncmp(argv[1], "random", length)) == 0) {
        if (argc < 2) {
            Tcl_SetResult(interp, "wrong number of arguments", TCL_STATIC);
            return TCL_ERROR;
        } else
            return (SdbRandomCmd(interp, sdbPtr, argc - 2, argv + 2));
    } else if ((c == 'c') && (strncmp(argv[1], "curve", length)) == 0) {
        if (argc < 3) {
            Tcl_SetResult(interp, "wrong number of arguments", TCL_STATIC);
            return TCL_ERROR;
        } else
            return (SdbCurves(interp, sdbPtr, argc - 2, &argv[2]));
    } else if ((c == 'a') && (strncmp(argv[1], "allocate", length)) == 0) {
        if (argc != 4) {
            Tcl_SetResult(interp, "wrong number of arguments", TCL_STATIC);
            return TCL_ERROR;
        } else {
            int                     flag;

            if (SdbFound(varPtr, argv[2])) {
                Tcl_AppendResult(interp, "field already exists \"", argv[2], "\"", NULL);
                return TCL_ERROR;
            }
            if (!strcmp(argv[3], "CODE"))
                flag = SDB_CODE | SDB_ALLOCATED | SDB_OUTPUT;
            else
                flag = SDB_NUMBER | SDB_ALLOCATED | SDB_OUTPUT;
            varPtr = SdbInsert(sdbPtr, argv[2], flag, sdbPtr->numRec);
            SdbReOptimize(sdbPtr);
        }
        Tcl_AppendResult(interp, argv[2], NULL);
        return TCL_OK;
    } else if ((c == 'p') && (strncmp(argv[1], "position", length) == 0)) {
        int                     pos;
        char                    str[20];

        if ((argc != 3) || (!sscanf(argv[2], "%d", &pos))) {
            Tcl_SetResult(interp, "Invalid record number", TCL_STATIC);
            return TCL_ERROR;
        }
        if (pos < 0)
            pos = 0;
        if (pos >= sdbPtr->numRec) {
            Tcl_SetResult(interp, "Invalid record number (> max)", TCL_STATIC);
            return TCL_ERROR;
        }
        sdbPtr->stamp++;
        sdbPtr->position = pos;
        sprintf(str, "%d", pos);
        Tcl_AppendResult(interp, str, NULL);
        return TCL_OK;
    } else if ((c == 's') && (strncmp(argv[1], "seek", length) == 0)) {
        int                     pos;
        char                    str[20];

        if (argc != 3) {
            Tcl_AppendResult(interp, "wrong # args: should be: \"",
                    argv[0], " seek ?id?\"", NULL);
            return TCL_ERROR;
        }
        pos = SdbSeek(interp, sdbPtr, argv[2]);
        if (pos < 0) {
            Tcl_AppendResult(interp, "could not find key \"", argv[2], "\"", NULL);
            return TCL_ERROR;
        }
        sdbPtr->stamp++;
        sdbPtr->position = pos;
        sprintf(str, "%d", pos);
        Tcl_AppendResult(interp, str, NULL);
        return TCL_OK;
    } else if (strcmp(argv[1], "index") == 0) {
        if (argc != 3) {
            Tcl_AppendResult(interp, "wrong # args: should be: \"",
                    argv[0], " index ?field?\"", NULL);
            return TCL_ERROR;
        }
        return (SdbIndex(interp, sdbPtr, argv[2]));
    } else if ((c == 'g') && (strncmp(argv[1], "get", length) == 0)) {
        if (argc != 3) {
            Tcl_AppendResult(interp, "wrong # args: should be \"",
                    argv[0], " get ?recnum or current?\"", (char *) NULL);
            return TCL_ERROR;
        }
        if (strcmp(argv[2], "current") == 0) {
            if (sdbPtr->position == sdbPtr->numRec) {
                Tcl_SetResult(interp, "end of file", TCL_STATIC);
                return TCL_ERROR;
            }
            errcode = SdbGetRecord(interp, sdbPtr, sdbPtr->position);
        } else {
            int                     rec;

            if ((sscanf(argv[2], "%d", &rec) != 1) ||
                    (rec < 0) || (rec >= sdbPtr->numRec)) {
                Tcl_AppendResult(interp, "invalid record number: ", argv[2],
                        (char *) NULL);
                return TCL_ERROR;
            }
            errcode = SdbGetRecord(interp, sdbPtr, rec);
        }
        return TCL_OK;
    } else if ((c == 'p') && (strncmp(argv[1], "put", length) == 0)) {
        int                     rec;

        if (argc != 4) {
            Tcl_AppendResult(interp, "wrong # args: should be \"",
                    argv[0], " put ?recnum or current? list\"", NULL);
            return TCL_ERROR;
        }
        if (strcmp(argv[2], "current") == 0) {
            /* do nothing */ ;
        } else {
            if ((sscanf(argv[2], "%d", &rec) != 1) ||
                    (rec < 0) || (rec >= sdbPtr->numRec)) {
                Tcl_AppendResult(interp, "invalid record number: ", argv[2],
                        (char *) NULL);
                return TCL_ERROR;
            }
        }
        return (SdbPutRecord(interp, sdbPtr, rec, argv[3]));
    } else if (strcmp(argv[1], "reload") == 0) {
        if (argc != 3) {
            Tcl_SetResult(interp, "wrong number of arguments", TCL_STATIC);
            return TCL_ERROR;
        } else {
            return (SdbReloadFile(interp, sdbPtr, argv[2]));
        }
    } else if (strcmp(argv[1], "bsave") == 0) {
        if (argc != 3) {
            Tcl_SetResult(interp, "wrong number of arguments", TCL_STATIC);
            return TCL_ERROR;
        } else {
            return (SdbBsave(interp, sdbPtr, argv[2]));
        }
    } else if (strcmp(argv[1], "bload") == 0) {
        if (argc != 3) {
            Tcl_SetResult(interp, "wrong number of arguments", TCL_STATIC);
            return TCL_ERROR;
        } else {
            return (SdbBload(interp, sdbPtr, argv[2]));
        }
    } else if (strcmp(argv[1], "snapshot") == 0) {
        if (argc != 3) {
            Tcl_SetResult(interp, "wrong number of arguments", TCL_STATIC);
            return TCL_ERROR;
        } else {
            return (SdbSnapshot(interp, sdbPtr, argv[2]));
        }
    } else if (strcmp(argv[1], "sort") == 0) {
        if (argc < 3) {
            Tcl_SetResult(interp, "wrong number of arguments", TCL_STATIC);
            return TCL_ERROR;
        } else {
            return (SdbSort(interp, sdbPtr, argc - 2, &argv[2]));
        }
    } else if (strcmp(argv[1], "summary") == 0) {
        if (argc < 4) {
            Tcl_SetResult(interp, "wrong number of arguments", TCL_STATIC);
            return TCL_ERROR;
        } else {
            return (SdbSummary(interp, sdbPtr, argc - 2, &argv[2]));
        }
    } else {
        Tcl_AppendResult(interp, "bad option \"", argv[1],
                "\": should be allocate, arrange, bload, bsave, calculate, close, curve, define, delete, format, get, index, info, join, position, put, random, reload, seek, select, snapshot, sort or summary",
                (char *) NULL);
        return TCL_ERROR;
    }
    return TCL_OK;
}

static void
SdbDeleteWidget(clientData)
     ClientData              clientData;
{

    /* this routine cleans up data structures and delete the memory
     * resident database */

    Tcl_HashEntry          *entryPtr;
    SdbFile                *sdbPtr = (SdbFile *) clientData;
    Tcl_Interp             *interp = sdbPtr->interp;

    /* withdraw the catalog entry */
    entryPtr = Tcl_FindHashEntry(&sdb_HashTable, sdbPtr->dbname);
    Tcl_DeleteHashEntry(entryPtr);

    /* Get rid of the variable associated with this database */
    Tcl_UntraceVar(interp, sdbPtr->dbname,
            TCL_TRACE_WRITES | TCL_TRACE_READS | TCL_GLOBAL_ONLY,
            SdbUpdate, (ClientData) sdbPtr->traceId);
    Tcl_UnsetVar(interp, sdbPtr->dbname, 0);

    /* Then call a routine to free storage allocated by the database */
    SdbCloseDB(sdbPtr);

}

/* ARGSUSED */
static int
SdbCmd(clientdata, interp, argc, argv)
     ClientData              clientdata;  /* Hash table pointer */
     Tcl_Interp             *interp;   /* Current interpreter. */
     int                     argc;     /* Number of arguments. */
     char                  **argv;     /* Argument strings. */
{
    unsigned int            length;
    char                    c;
    SdbFile                *sdbPtr;
    SdbItemData            *id;
    int                     new;
    Tcl_HashEntry          *entryPtr;

    if (argc < 3) {
        Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
                " option dbname ?arg arg ...?\"", (char *) NULL);
        return TCL_ERROR;
    }
    c = argv[1][0];
    length = strlen(argv[1]);

    entryPtr = Tcl_FindHashEntry(&sdb_HashTable, argv[2]);
    if (entryPtr) {
        Tcl_AppendResult(interp, "database \"", argv[2], "\" already exists", NULL);
        return TCL_ERROR;
    }
    if ((c == 'o') && (strncmp(argv[1], "open", length) == 0)) {
        char                   *file,
                               *name;

        if (argc != 4) {
            Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
                    " open dbname file\"", (char *) NULL);
            return TCL_ERROR;
        }
        name = argv[2];
        file = argv[3];

        sdbPtr = (SdbFile *) SdbOpenFile(interp, file, name, NULL);

        if (!sdbPtr) {
            Tcl_AppendResult(interp, "\nfailed to open database ", name, NULL);
            return TCL_ERROR;
        }
        sdbPtr->filename = (char *) ckalloc(strlen(file) + 1);
        strcpy(sdbPtr->filename, file);

    } else if (strcmp(argv[1], "create") == 0) {

        int                     nrecs;

        if (argc != 4) {
            Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
                    " create dbname nrecs\"", (char *) NULL);
            return TCL_ERROR;
        }
        if (sscanf(argv[3], "%d", &nrecs) != 1) {
            Tcl_AppendResult(interp, "invalid record count \"", argv[3],
                    "\"", (char *) NULL);
            return TCL_ERROR;
        }
        sdbPtr = SdbCreateDB(interp, argv[2], nrecs);

        if (!sdbPtr) {
            Tcl_AppendResult(interp, "\nfailed to create database ", argv[2], NULL);
            return TCL_ERROR;
        }
    } else {
        Tcl_AppendResult(interp, "bad option \"", argv[1],
                "\": should be create or open",
                (char *) NULL);
        return TCL_ERROR;
    }

    sdbPtr->interp = interp;
    Tcl_CreateCommand(interp, argv[2], SdbWidget, (ClientData) sdbPtr,
            SdbDeleteWidget);

    id = (SdbItemData *) ckalloc(sizeof(SdbItemData));

    id->varPtr = NULL;
    id->dbPtr = sdbPtr;
    sdbPtr->traceId = id;
    Tcl_TraceVar(sdbPtr->interp, sdbPtr->dbname,
            TCL_TRACE_WRITES | TCL_TRACE_READS | TCL_GLOBAL_ONLY,
            SdbUpdate, (ClientData) id);

    entryPtr = Tcl_CreateHashEntry(&sdb_HashTable, argv[2], &new);
    Tcl_SetHashValue(entryPtr, sdbPtr);

    Tcl_SetResult(interp, argv[2], TCL_VOLATILE);
    return TCL_OK;
}

int
SdbSnapshot(interp, sdbPtr, fileName)
     Tcl_Interp             *interp;
     SdbFile                *sdbPtr;
     char                   *fileName;
{

    int                     j,
                            rec;
    SdbVariable            *varPtr = sdbPtr->varPtr;
    FILE                   *output;
    _bit_type              *selectPtr = sdbPtr->select;

    /* open the output file */
    if (!fileName) {
        Tcl_AppendResult(interp, "invalid file name", NULL);
        return TCL_ERROR;
    }
    output = fopen(fileName, "w");
    if (!output) {
        Tcl_AppendResult(interp, "could not open output file \"",
                fileName, "\"", NULL);
        return TCL_ERROR;
    }
    /* print out the variable header record */
    Tcl_ResetResult(interp);
    Tcl_AppendElement(interp, "VAR");
    for (varPtr = sdbPtr->varPtr; varPtr; varPtr = varPtr->next) {
        if (varPtr->id && varPtr->flag & SDB_OUTPUT) {
            if (varPtr->flag & SDB_CODE)
                Tcl_AppendElement(interp, varPtr->id);
        }
    }
    Tcl_AppendElement(interp, ":");
    for (varPtr = sdbPtr->varPtr; varPtr; varPtr = varPtr->next) {
        if (varPtr->id && varPtr->flag & SDB_OUTPUT) {
            if (varPtr->flag & SDB_NUMBER)
                Tcl_AppendElement(interp, varPtr->id);
        }
    }
    fprintf(output, "%s\n", interp->result);
    Tcl_ResetResult(interp);

    /* loop through and print out all selected records */
    for (j = 0; j < sdbPtr->numRec; j++) {
        rec = sdbPtr->base[j];
        if (bit_test(selectPtr, rec)) {
            sdbPtr->stamp++;
            Tcl_AppendElement(interp, "DATA");
            for (varPtr = sdbPtr->varPtr; varPtr; varPtr = varPtr->next) {
                if (varPtr->flag & SDB_OUTPUT) {
                    if (varPtr->flag & SDB_CODE)
                        Tcl_AppendElement(interp, SdbGetCode(sdbPtr, varPtr, rec));
                }
            }
            Tcl_AppendElement(interp, ":");
            for (varPtr = sdbPtr->varPtr; varPtr; varPtr = varPtr->next) {
                if (varPtr->flag & SDB_OUTPUT) {
                    if (varPtr->flag & SDB_NUMBER) {
                        char                    temp[32];

                        sprintf(temp, "%-.15g", SdbGetNumber(sdbPtr, varPtr, rec));
                        Tcl_AppendElement(interp, temp);
                    }
                }
            }
            fprintf(output, "%s\n", interp->result);
            Tcl_ResetResult(interp);
        }
    }
    fclose(output);
    sdbPtr->stamp++;
    return TCL_OK;
}

static int
SdbPutRecord(interp, sdbPtr, rec, list)
     Tcl_Interp             *interp;
     SdbFile                *sdbPtr;
     int                     rec;
     char                   *list;
{

    int                     listArgc,
                            item_count,
                            j,
                            position;
    char                  **listArgv;
    SdbVariable            *varPtr;

    sdbPtr->position = rec;
    sdbPtr->stamp++;
    position = sdbPtr->base[rec];

    if (!list) {
        Tcl_AppendResult(interp, "no values in list", NULL);
        return TCL_ERROR;
    }
    if (Tcl_SplitList(interp, list, &listArgc, &listArgv) != TCL_OK) {
        Tcl_AppendResult(interp, "invalid list", NULL);
        return TCL_ERROR;
    }
    item_count = 0;
    for (varPtr = sdbPtr->varPtr; varPtr; varPtr = varPtr->next) {
        if (varPtr->flag & SDB_OUTPUT) {
            item_count++;
        }
    }

    if (listArgc != item_count) {
        Tcl_AppendResult(interp, "wrong number of items in list", NULL);
        ckfree(listArgv);
        return TCL_ERROR;
    }
    j = 0;
    for (varPtr = sdbPtr->varPtr; varPtr; varPtr = varPtr->next) {
        if (varPtr->flag & SDB_OUTPUT) {
            if (varPtr->flag & (SDB_CONSTANT | SDB_ALLOCATED | SDB_LOOKUP)) {
                if (varPtr->flag & SDB_CODE)
                    SdbPutCode(sdbPtr, varPtr, listArgv[j], position);
                else {
                    double                  value;

                    if (sscanf(listArgv[j], "%lf", &value) != 1)
                        value = 0.0;
                    SdbPutNumber(sdbPtr, varPtr, value, position);
                }
            }
            j++;
        }
    }
    ckfree(listArgv);
    return TCL_OK;
}

double
SdbPutNumber(sdbPtr, varPtr, value, position)
     SdbFile                *sdbPtr;
     SdbVariable            *varPtr;
     double                  value;
     int                     position;
{
    if (varPtr->flag & SDB_CONSTANT)
        varPtr->u.value = value;
    else if (varPtr->flag & SDB_ALLOCATED)
        varPtr->u.dvalue[position] = value;
    else if (varPtr->flag & SDB_LOOKUP)
        SdbPutNumberLookup(sdbPtr, (varPtr->u.expr[0])->content.luHandlePtr,
                ((varPtr->u.expr[0])->right)->content.integer,
                position, value);
    return value;
}

char                   *
SdbPutCode(sdbPtr, varPtr, code, position)
     SdbFile                *sdbPtr;
     SdbVariable            *varPtr;
     char                   *code;
     int                     position;
{
    int                     new;

    /* place this string into the hash table for safe keeping */
    code = GetIndices(sdbPtr->htbl, code, &new);
    if (varPtr->flag & SDB_CONSTANT)
        varPtr->u.code = code;
    else if (varPtr->flag & SDB_ALLOCATED)
        varPtr->u.ivalue[position] = code;
    else if (varPtr->flag & SDB_LOOKUP)
        SdbPutCodeLookup(sdbPtr, (varPtr->u.expr[0])->content.luHandlePtr,
                ((varPtr->u.expr[0])->right)->content.integer,
                position, code);
    return code;
}

static int
SdbGetRecord(interp, sdbPtr, rec)
     Tcl_Interp             *interp;
     SdbFile                *sdbPtr;
     int                     rec;
{

    SdbVariable            *varPtr;
    char                    temp[256];

    sdbPtr->position = rec;
    rec = sdbPtr->base[rec];
    sdbPtr->stamp++;

    for (varPtr = sdbPtr->varPtr; varPtr; varPtr = varPtr->next) {
        if (varPtr->flag & SDB_OUTPUT) {
            if (varPtr->flag & SDB_CODE) {
                sprintf(temp, varPtr->format, SdbGetCode(sdbPtr, varPtr, rec));
                Tcl_AppendElement(interp, temp);
            } else if (varPtr->flag & SDB_NUMBER) {
                sprintf(temp, varPtr->format, SdbGetNumber(sdbPtr, varPtr, rec));
                Tcl_AppendElement(interp, temp);
            }
        }
    }
    return TCL_OK;
}

char                   *
SdbGetCode(sdbPtr, varPtr, rec)
     SdbFile                *sdbPtr;
     SdbVariable            *varPtr;
     int                     rec;
{
    static char            *missing = ".";

    if (varPtr->flag & SDB_CODE) {
        if (varPtr->flag & SDB_CONSTANT)
            return (varPtr->u.code);
        else if (varPtr->flag & SDB_ALLOCATED)
            return (varPtr->u.ivalue[rec]);
        else {
            if (sdbPtr->stamp != varPtr->stamp) {
                if (varPtr->flag & (SDB_EXPRESSION | SDB_LOOKUP)) {
                    varPtr->cache.code =
                            SdbCharEval(sdbPtr, rec, varPtr->u.expr[0]);
                } else {
                    fprintf(stderr, "bad variable type\n");
                    exit(1);
                }
                varPtr->stamp = sdbPtr->stamp;
            }
            return (varPtr->cache.code);
        }
    }
    return (missing);
}

double
SdbGetNumber(sdbPtr, varPtr, rec)
     SdbFile                *sdbPtr;
     SdbVariable            *varPtr;
     int                     rec;
{

    if (varPtr->flag & SDB_NUMBER) {
        if (varPtr->flag & SDB_CONSTANT)
            return (varPtr->u.value);
        else if (varPtr->flag & SDB_ALLOCATED)
            return (varPtr->u.dvalue[rec]);
        else {
            if (sdbPtr->stamp != varPtr->stamp) {
                if (varPtr->flag & (SDB_EXPRESSION | SDB_LOOKUP)) {
                    varPtr->cache.value =
                            SdbEvaluate(sdbPtr, rec, varPtr->u.expr[0]);
                } else {
                    fprintf(stderr, "bad variable type\n");
                    exit(1);
                }
                varPtr->stamp = sdbPtr->stamp;
            }
            return (varPtr->cache.value);
        }
    }
    return (0.0);
}

static int
SdbGetCodeList(interp, sdbPtr, varPtr)
     Tcl_Interp             *interp;
     SdbFile                *sdbPtr;
     SdbVariable            *varPtr;
{

    /* build up a list of unique codes in a field */
    int                     rec,
                            new;
    _bit_type              *selectPtr = sdbPtr->select;
    char                   *code;
    Tcl_HashTable           tempTable;
    Tcl_HashEntry          *entryPtr;
    Tcl_HashSearch          search;

    /* first scan the database and store unique codes in a hash table */
    Tcl_InitHashTable(&tempTable, TCL_STRING_KEYS);
    for (rec = 0; rec < sdbPtr->numRec; rec++) {
        if (bit_test(selectPtr, rec)) {
            sdbPtr->stamp++;
            code = SdbGetCode(sdbPtr, varPtr, rec);
            entryPtr = Tcl_CreateHashEntry(&tempTable, code, &new);
        }
    }

    /* then search the hash table and append the unique codes to
     * the result string */
    for (entryPtr = Tcl_FirstHashEntry(&tempTable, &search);
            entryPtr != NULL;
            entryPtr = Tcl_NextHashEntry(&search)) {
        Tcl_AppendElement(interp, entryPtr->key.string);
    }

    /* clean up and return */
    Tcl_DeleteHashTable(&tempTable);
    return TCL_OK;
}
static int
SdbGetNumberRange(interp, sdbPtr, varPtr)
     Tcl_Interp             *interp;
     SdbFile                *sdbPtr;
     SdbVariable            *varPtr;
{

    /* find min and max values in a field */
    int                     rec;
    _bit_type              *selectPtr = sdbPtr->select;
    double                  min = 1e100,
                            max = -1e100,
                            value;
    char                    temp[64];

    /* first scan the database and find extremes */
    for (rec = 0; rec < sdbPtr->numRec; rec++) {
        if (bit_test(selectPtr, rec)) {
            sdbPtr->stamp++;
            value = SdbGetNumber(sdbPtr, varPtr, rec);
            if (value < min)
                min = value;
            if (value > max)
                max = value;
        }
    }

    sprintf(temp, "%-.15g %-.15g", min, max);
    Tcl_AppendResult(interp, temp, NULL);

    return TCL_OK;
}

int
SdbRandomCmd(interp, dbPtr, argc, argv)
     Tcl_Interp             *interp;
     SdbFile                *dbPtr;
     int                     argc;
     char                  **argv;
{
    Tcl_HashEntry          *entryPtr;
    Tcl_HashSearch          search;
    SdbRandom              *randomPtr;

    if (argc == 1 && !strcmp("list", argv[0])) {
        for (entryPtr = Tcl_FirstHashEntry(dbPtr->randomList, &search);
                entryPtr != NULL;
                entryPtr = Tcl_NextHashEntry(&search))
            Tcl_AppendElement(interp, Tcl_GetHashKey(dbPtr->randomList, 
						     entryPtr));
      return TCL_OK;
      }

    if (argc < 2) {
        Tcl_AppendResult(interp, "wrong number of arguments", NULL);
        return TCL_ERROR;
    }
    entryPtr = Tcl_FindHashEntry(dbPtr->randomList, argv[1]);

    if (!strcmp("info", argv[0])) {
      if (!entryPtr) {
	Tcl_AppendResult(interp,"random stream \"",argv[1],"\" unknnown",
			 NULL);
	return TCL_ERROR;
      }
      randomPtr = (SdbRandom *) Tcl_GetHashValue(entryPtr);
      if (randomPtr->type == uniform) 
	Tcl_AppendResult(interp,"uniform",NULL);
      else if (randomPtr->type == normal)
	Tcl_AppendResult(interp,"normal",NULL);
      else
	Tcl_AppendResult(interp,"unknown",NULL);
      return TCL_OK;
    }
    else if (!strcmp("seed", argv[0])) {
        int                     seed;

        if (argc != 3) {
            Tcl_AppendResult(interp, "wrong # args:  seed name value", NULL);
            return TCL_ERROR;
        }
        if (Tcl_GetInt(interp, argv[2], &seed) != TCL_OK)
            return TCL_ERROR;
        if (!entryPtr)
            randomPtr = SdbCreateRandom(dbPtr, argv[1]);
        else
            randomPtr = (SdbRandom *) Tcl_GetHashValue(entryPtr);
        initstate((unsigned) seed, randomPtr->state, 128);
	Tcl_AppendResult(interp,argv[1],NULL);
        return TCL_OK;
    }
    else if (!strcmp("type", argv[0])) {
      if (argc != 3) {
	Tcl_AppendResult(interp, "wrong # args:  type name distribution", NULL);
	return TCL_ERROR;
        }
      if (!entryPtr)
	randomPtr = SdbCreateRandom(dbPtr, argv[1]);
      else
	randomPtr = (SdbRandom *) Tcl_GetHashValue(entryPtr);
      if (!strcmp("uniform",argv[2])) 
	randomPtr->type = uniform;
      else if (!strcmp("normal",argv[2])) 
	randomPtr->type = normal;
      else {
	Tcl_AppendResult(interp, "invalid distribution \"", argv[2], 
			 "\"", NULL);
	return TCL_ERROR;
      }
      Tcl_AppendResult(interp,argv[1],NULL);
      return TCL_OK;
    }     
    else {
      Tcl_AppendResult(interp, "invalid argument \"", argv[0], 
		       "\", should be info, list, seed or type", NULL);
      return TCL_ERROR;
    }
    return TCL_OK;

}

SdbRandom              *
SdbCreateRandom(dbPtr, sequence)
     SdbFile                *dbPtr;
     char                   *sequence;
{
    int                     new;
    SdbRandom              *randomPtr;
    Tcl_HashEntry          *entryPtr;

    entryPtr = Tcl_CreateHashEntry(dbPtr->randomList, sequence, &new);
    randomPtr = (SdbRandom *) ckalloc(sizeof(SdbRandom));

    randomPtr->state = (char *) ckalloc(128);
    initstate(1, randomPtr->state, 128);
    randomPtr->type = uniform;
    Tcl_SetHashValue(entryPtr, randomPtr);
    return randomPtr;
}
